package types

import (
	"crypto/ecdsa"
	"fmt"
	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/common/hexutil"
	"github.com/ethereum/go-ethereum/common/math"
	"github.com/ethereum/go-ethereum/core/types"
	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ethereum/go-ethereum/signer/core/apitypes"
	"math/big"
)

type Account struct {
	Address    string `json:"address"`
	PrivateKey string `json:"private"`
	//Nonce      uint64            `json:"-"`
	PK        *ecdsa.PrivateKey `json:"-"`
	Addr      common.Address    `json:"-"`
	Recipient string            `json:“recipient"`
}

func (a *Account) sign(chainid *big.Int, tx *types.Transaction) (*types.Transaction, error) {
	signer := types.NewEIP155Signer(chainid)
	h := signer.Hash(tx)
	sig, err := crypto.Sign(h[:], a.PK)
	if err != nil {
		return nil, err
	}
	return tx.WithSignature(signer, sig)
}

func (a *Account) SignTx(addr common.Address, tx *types.Transaction) (*types.Transaction, error) {
	return a.sign(big.NewInt(56), tx)
}

func (a *Account) SignMessage(msg []byte) ([]byte, error) {
	sig, err := crypto.Sign(msg, a.PK)
	if err != nil {
		return nil, err
	}
	return sig, nil
}

func (a *Account) SignTypedData(chainID *big.Int, permit2Address, tokenAddress common.Address,
	amount *big.Int, expiration *big.Int, nonce *big.Int) ([]byte, error) {

	mchainId := math.HexOrDecimal256(*chainID)

	// 创建EIP-712结构化数据
	typedData := apitypes.TypedData{
		Types: apitypes.Types{
			"EIP712Domain": []apitypes.Type{
				{Name: "name", Type: "string"},
				{Name: "version", Type: "string"},
				{Name: "chainId", Type: "uint256"},
				{Name: "verifyingContract", Type: "address"},
			},
			"TokenPermissions": []apitypes.Type{
				{Name: "token", Type: "address"},
				{Name: "amount", Type: "uint256"},
			},
			"PermitTransferFrom": []apitypes.Type{
				{Name: "permitted", Type: "TokenPermissions"},
				{Name: "nonce", Type: "uint256"},
				{Name: "deadline", Type: "uint256"},
			},
		},
		PrimaryType: "PermitTransferFrom",
		Domain: apitypes.TypedDataDomain{
			Name:              "Permit2",
			Version:           "1",
			ChainId:           &mchainId,
			VerifyingContract: permit2Address.Hex(),
		},
		Message: apitypes.TypedDataMessage{
			"permitted": map[string]interface{}{
				"token":  tokenAddress.Hex(),
				"amount": amount.String(),
			},
			"nonce":    hexutil.EncodeUint64(nonce.Uint64()),
			"deadline": expiration.String(),
		},
	}

	domain, err := typedData.HashStruct("EIP712Domain", typedData.Domain.Map())
	if err != nil {
		return nil, fmt.Errorf("domain hash failed: %v", err)
	}

	permitHash, err := typedData.HashStruct(typedData.PrimaryType, typedData.Message)
	if err != nil {
		return nil, fmt.Errorf("message hash failed: %v", err)
	}

	rawData := []byte(fmt.Sprintf("\x19\x01%s%s", string(domain), string(permitHash)))
	signHash := crypto.Keccak256(rawData)

	signature, err := crypto.Sign(signHash, a.PK)
	if err != nil {
		return nil, fmt.Errorf("sign message failed: %v", err)
	}

	signature[64] += 27

	return signature, nil
}
